/* Copyright (C) 2013 Bjarke Hammersholt Roune (www.broune.com)
   MemTailor is distributed under the Modified BSD License. See license.txt. */

#include "memtailor.h"

extern "C" {
  void libmemtailorIsPresent(void) {}
  char MEMTAILOR_VERSION_STRING[] = "1.0";
}
